/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed_internal.h"

const char* addr2char(char ip[IPCHAR_SIZE], const address_t *a)
{
  if(a->a.sockaddr.sa_family == AF_INET)
  {
    const uint8_t *s4 = (const uint8_t*)&a->a.sockaddr_in.sin_addr.s_addr;
    (void)snprintf(ip, IPCHAR_SIZE, "%u.%u.%u.%u", (unsigned int)s4[0], (unsigned int)s4[1], (unsigned int)s4[2], (unsigned int)s4[3]);
    ip[IPCHAR_SIZE-1] = '\0';
  }
  else if(a->a.sockaddr.sa_family == AF_INET6)
  {
    int error = getnameinfo((const struct sockaddr*)a, (socklen_t)sizeof(*a), ip, IPCHAR_SIZE, NULL, 0, NI_NUMERICHOST);
    if(unlikely(error != 0))
    {
      ip[0] = '\0';
    }
  }
  else
  {
    return "unknown";
  }
  return ip;
}

int char2addr(const char *ip, in_port_t port, address_t *addr)
{
  int i, dot=0, colon=0, hex=0;
  struct addrinfo *res = NULL, hints;

  if(unlikely(ip == NULL)) return -1;

  for(i = 0; ip[i] != '\0'; i++)
  {
    if(ip[i] >= '0' && ip[i] <= '9') continue;
    if(ip[i] == '.')
    {
      dot++;
    }
    else if(ip[i] == ':')
    {
      colon++;
    }
    else if((ip[i] >= 'a' && ip[i] <= 'f') || (ip[i] >= 'A' && ip[i] <= 'F'))
    {
      hex++;
    }
    else
    {
      return -1;
    }
  }
  if(unlikely(hex != 0 && colon == 0)) return -1;
  if(dot != 3)
  {
    if(unlikely(dot != 0 || colon < 2 || colon > 8)) return -1;
  }
  else
  {
    if(unlikely(colon == 1 || colon > 6)) return -1;
  }

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  if(unlikely(getaddrinfo(ip, NULL, &hints, &res) != 0)) return -1;
  if(unlikely(res == NULL)) return -1;
  if(unlikely(res->ai_addrlen > (socklen_t)sizeof(*addr)))
  {
    freeaddrinfo(res);
    return -1;
  }
  if(addr != NULL)
  {
    memset(addr, 0, sizeof(*addr));
    memcpy(addr, res->ai_addr, (size_t)res->ai_addrlen);

    if(res->ai_family == AF_INET)
    {
      addr->a.sockaddr_in.sin_port = (in_port_t)htons((in_port_t)port);
    }
    else if(res->ai_family == AF_INET6)
    {
      addr->a.sockaddr_in6.sin6_port = (in_port_t)htons((in_port_t)port);
    }
  }
  freeaddrinfo(res);

  return 0;
}

__attribute__((pure)) int addr_cmp(const address_t *a, const address_t *b)
{
  if(a->a.sockaddr.sa_family != b->a.sockaddr.sa_family) return -1;
  if(a->a.sockaddr.sa_family == AF_INET)
  {
    uint32_t au = ntohl(a->a.sockaddr_in.sin_addr.s_addr);
    uint32_t bu = ntohl(a->a.sockaddr_in.sin_addr.s_addr);
    if(au < bu) return -1;
    if(au > bu) return 1;
    return 0;
  }
  if(a->a.sockaddr.sa_family == AF_INET6)
  {
    return memcmp(&a->a.sockaddr_in6.sin6_addr.s6_addr, &b->a.sockaddr_in6.sin6_addr.s6_addr, 16);
  }
  return memcmp(a, b, sizeof(*a));
}

__attribute__((pure)) int is_in_range(const address_t *address, const ip_range_t *range)
{
  if(address->a.sockaddr.sa_family != range->begin.a.sockaddr.sa_family || address->a.sockaddr.sa_family != range->end.a.sockaddr.sa_family) return 0;

  if(address->a.sockaddr.sa_family == AF_INET)
  {
    uint32_t a = ntohl(address->a.sockaddr_in.sin_addr.s_addr);
    uint32_t b = ntohl(range->begin.a.sockaddr_in.sin_addr.s_addr);
    uint32_t e = ntohl(range->end.a.sockaddr_in.sin_addr.s_addr);
    if(a < b || a > e) return 0;
    return 1;
  }
  if(address->a.sockaddr.sa_family == AF_INET6)
  {
    int i;
    for(i = 0; i < 16; i++)
    {
      if(address->a.sockaddr_in6.sin6_addr.s6_addr[i] < range->begin.a.sockaddr_in6.sin6_addr.s6_addr[i]) return 0;
      if(address->a.sockaddr_in6.sin6_addr.s6_addr[i] > range->begin.a.sockaddr_in6.sin6_addr.s6_addr[i]) break;
    }
    for(i = 0; i < 16; i++)
    {
      if(address->a.sockaddr_in6.sin6_addr.s6_addr[i] > range->end.a.sockaddr_in6.sin6_addr.s6_addr[i]) return 0;
      if(address->a.sockaddr_in6.sin6_addr.s6_addr[i] < range->end.a.sockaddr_in6.sin6_addr.s6_addr[i]) break;
    }
    return 1;
  }
  return 0;
}

static int hex_digit(const char *buffer, uint8_t *value)
{
  *value = 0;
  if(buffer[0] >= '0' && buffer[0] <= '9')
    *value = (uint8_t)(buffer[0] - '0');
  else if(buffer[0] >= 'A' && buffer[0] <= 'F')
    *value = (uint8_t)((buffer[0] - 'A') + (char)10);
  else if(buffer[0] >= 'a' && buffer[0] <= 'f')
    *value = (uint8_t)((buffer[0] - 'a') + (char)10);
  else
    return 0;
  if(buffer[1] >= '0' && buffer[1] <= '9')
    *value = (uint8_t)((*value << 4) | (buffer[1] - '0'));
  else if(buffer[0] >= 'A' && buffer[0] <= 'F')
    *value = (uint8_t)((*value << 4) | ((buffer[1] - 'A') + (char)10));
  else if(buffer[0] >= 'a' && buffer[0] <= 'f')
    *value = (uint8_t)((*value << 4) | ((buffer[0] - 'a') + (char)10));
  else
    return 1;
  return 2;
}

static int dec_digit(const char *buffer, uint8_t *value)
{
  int v;
  *value = 0;
  if(buffer[0] < '0' || buffer[0] > '9') return 0;
  v = (int)(buffer[0] - '0');
  if(buffer[1] < '0' || buffer[1] > '9')
  {
    *value = (uint8_t)v;
    return 1;
  }
  v = v * 10 + (int)(buffer[1] - '0');
  if(buffer[2] < '0' || buffer[2] > '9')
  {
    *value = (uint8_t)v;
    return 2;
  }
  v = v * 10 + (int)(buffer[2] - '0');
  if(v > 255) return 0;
  *value = (uint8_t)v;
  return 3;
}

int addr6_prefix(const buffer_t *name, address_t *a)
{
  int i, pos = 0;
  memset(a, 0, sizeof(*a));
  a->a.sockaddr_in6.sin6_family = AF_INET6;
  for(i = 15; i >= 0; i--)
  {
    uint8_t value = 0;
    int inc = hex_digit((char*)name->value + pos, &value);
    if(inc == 0 || name->value[pos+inc] != (uint8_t)'.') break;
    a->a.sockaddr_in6.sin6_addr.s6_addr[i] = value;
    pos += inc + 1;
  }
  return 16 - (i + 1);
}

int addr_prefix(const buffer_t *name, address_t *a)
{
  int i, pos = 0;
  uint8_t *ad = (uint8_t*)&a->a.sockaddr_in.sin_addr.s_addr;
  memset(a, 0, sizeof(*a));
  a->a.sockaddr_in.sin_family = AF_INET;
  for(i = 3; i >= 0; i--)
  {
    uint8_t value = 0;
    int inc = dec_digit((char*)name->value + pos, &value);
    if(inc == 0 || name->value[pos+inc] != (uint8_t)'.') break;
    ad[i] = value;
    pos += inc + 1;
  }
  return 4 - (i + 1);
}

__attribute__((pure)) int is_subdomain(const char *domain, const char *check)
{
  int domain_length = (int)strlen(domain);
  int i, diff = (int)strlen(check) - domain_length;
  if(diff < 0) return 0;
  for(i = 0; i < domain_length; i++)
  {
    if(tolower((int)domain[i]) != tolower((int)check[diff + i])) return 0;
  }
  if(diff == 0) return 1;
  if(check[diff - 1] == '.') return 1;
  return 0;
}

int ptr_name(buffer_t *name, const address_t *a)
{
  if(a->a.sockaddr.sa_family == AF_INET)
  {
    const unsigned char *ad = (const unsigned char*)&a->a.sockaddr_in.sin_addr.s_addr;
    return buffer_set(name, "%u.%u.%u.%u.in-addr.arpa", ad[3], ad[2], ad[1], ad[0]);
  }
  else if(a->a.sockaddr.sa_family == AF_INET6)
  {
    const unsigned char *ad = (const unsigned char *)a->a.sockaddr_in6.sin6_addr.s6_addr;
    return buffer_set(name, "%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.%02x.ip6.arpa",
      ad[15], ad[14], ad[13], ad[12], ad[11], ad[10], ad[9], ad[8], ad[7], ad[6], ad[5], ad[4], ad[3], ad[2], ad[1], ad[0]);
  }
  return -1;
}

/*
int my_addr(address_t *src, const address_t *dst, in_port_t port)
{
  SOCKET fd = socket((int)dst->a.sockaddr.sa_family, SOCK_DGRAM, 0);
  if(unlikely(fd == INVALID_SOCKET)) return -1;
  if(likely(connect(fd, &dst->a.sockaddr, (socklen_t)sizeof(*dst)) == 0))
  {
    socklen_t len = (socklen_t)sizeof(*src);
    if(likely(getsockname(fd, &src->a.sockaddr, &len) == 0))
    {
      (void)closesocket(fd);
      if(likely(src->a.sockaddr.sa_family == AF_INET))
      {
        src->a.sockaddr_in.sin_port = (in_port_t)htons((in_port_t)port);
      }
      else if(likely(src->a.sockaddr.sa_family == AF_INET6))
      {
        src->a.sockaddr_in6.sin6_port = (in_port_t)htons((in_port_t)port);
      }
      return 0;
    }
  }
  (void)closesocket(fd);
  return -1;
}
*/
